@@ -10,6 +10,7 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             this.usedTerms = [];
             this.completer = null;
             this.lastCompletedTerm = null;
+            this.manageRequired = input.required;
             this._dataInput = null;
             this._termInput = null;
             this._termContainer = null;
@@ -147,7 +148,15 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                 this.togglePlaceholder();
             }
 
-            return this.hasTerms();
+            if (this.hasTerms()) {
+                if (this.manageRequired) {
+                    this.input.required = false;
+                }
+
+                return true;
+            }
+
+            return false;
         }
 
         registerTerms() {
@@ -159,6 +168,10 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                     termData['class'] = label.className;
                 }
 
+                if (label.title) {
+                    termData['title'] = label.title;
+                }
+
                 this.registerTerm(this.decodeTerm(termData), label.dataset.index);
             });
         }
@@ -176,7 +189,22 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             // Reset the data input, otherwise the value remains and is sent continuously with subsequent requests
             this.dataInput.value = '';
 
-            for (const termIndex of Object.keys(changedTerms)) {
+            if (changedTerms === 'bogus') {
+                return;
+            }
+
+            let changedIndices = Object.keys(changedTerms);
+            if (! changedIndices.length) {
+                // Perform a partial reset. this.reset() empties the termContainer, which isn't desired here
+                this.usedTerms = [];
+                this.lastCompletedTerm = null;
+
+                this.registerTerms();
+                this.togglePlaceholder();
+                this.termInput.value = '';
+            }
+
+            for (const termIndex of changedIndices) {
                 let label = this.termContainer.querySelector(`[data-index="${ termIndex }"]`);
                 if (! label) {
                     continue;
@@ -213,7 +241,7 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
 
         readFullTerm(input, termIndex = null) {
             let value = this.readPartialTerm(input);
-            if (! value) {
+            if (! value && this.lastCompletedTerm === null) {
                 return false;
             }
 
@@ -223,11 +251,19 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                 termData = { ...this.usedTerms[termIndex] };
             }
 
-            termData.label = value;
-            termData.search = value;
+            if (value) {
+                termData.label = value;
+                termData.search = value;
+            }
 
             if (this.lastCompletedTerm !== null) {
-                if (termData.label === this.lastCompletedTerm.label) {
+                if ('type' in this.lastCompletedTerm && this.lastCompletedTerm.type === 'terms') {
+                    if (typeof this.lastCompletedTerm.terms === 'string') {
+                        termData = JSON.parse(this.lastCompletedTerm.terms);
+                    } else {
+                        termData = this.lastCompletedTerm.terms;
+                    }
+                } else if (termData.label === this.lastCompletedTerm.label) {
                     Object.assign(termData, this.lastCompletedTerm);
                 }
 
@@ -313,18 +349,55 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             return this.termsToQueryString(this.usedTerms);
         }
 
-        saveTerm(input, updateDOM = true) {
+        checkValidity(input) {
+            if (input.pattern && ! input.checkValidity()) {
+                if (! input.value.match(input.pattern)) {
+                    if (input.dataset.invalidMsg) {
+                        input.setCustomValidity(input.dataset.invalidMsg);
+                    }
+
+                    return false;
+                }
+
+                // If the pattern matches, reset the custom validity, otherwise the value is still invalid.
+                input.setCustomValidity('');
+            }
+
+            // The pattern isn't set or it matches. Any other custom validity must not be accounted for here.
+            return true;
+        }
+
+        reportValidity(element) {
+            setTimeout(() => element.reportValidity(), 0);
+        }
+
+        validate(element) {
+            if (! this.checkValidity(element)) {
+                this.reportValidity(element);
+
+                return false;
+            }
+
+            return true;
+        }
+
+        saveTerm(input, updateDOM = true, force = false) {
+            if (! this.checkValidity(input)) {
+                return false;
+            }
+
             let termIndex = input.parentNode.dataset.index;
             let termData = this.readFullTerm(input, termIndex);
 
-            // Only save if something has changed
+            // Only save if something has changed, unless forced
             if (termData === false) {
-                return this.removeTerm(input.parentNode, updateDOM);
-            } else if (this.usedTerms[termIndex].label !== termData.label) {
+                console.warn('[BaseInput] Input is empty, cannot save');
+            } else if (force || this.usedTerms[termIndex].label !== termData.label) {
+                let oldTermData = this.usedTerms[termIndex];
                 this.usedTerms[termIndex] = termData;
                 this.updateTermData(termData, input);
 
-                return termData;
+                return oldTermData;
             }
 
             return false;
@@ -337,6 +410,24 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             if (!! termData.search || termData.search === '') {
                 label.dataset.search = termData.search;
             }
+
+            if (!! termData.title) {
+                label.title = termData.title;
+            } else {
+                label.title = '';
+            }
+
+            if (termData.pattern) {
+                input.pattern = termData.pattern;
+                delete termData.pattern;
+
+                if (termData.invalidMsg) {
+                    input.dataset.invalidMsg = termData.invalidMsg;
+                    delete termData.invalidMsg;
+                }
+
+                this.validate(input);
+            }
         }
 
         termsToQueryString(terms) {
@@ -450,6 +541,10 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
         }
 
         togglePlaceholder() {
+            if (this.isTermDirectionVertical()) {
+                return;
+            }
+
             let placeholder = '';
 
             if (! this.hasTerms()) {
@@ -474,6 +569,10 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                 label.classList.add(termData.class);
             }
 
+            if (termData.title) {
+                label.title = termData.title;
+            }
+
             label.dataset.label = termData.label;
             label.dataset.search = termData.search;
             label.dataset.index = termIndex;
@@ -500,19 +599,38 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             return 'noAutoSubmit' in this.input.dataset;
         }
 
-        autoSubmit(input, changeType, changedTerms) {
-            if (this.shouldNotAutoSubmit()) {
+        shouldNotAutoSubmitOnRemove() {
+            return 'noAutoSubmitOnRemove' in this.input.dataset;
+        }
+
+        autoSubmit(input, changeType, data) {
+            if (this.shouldNotAutoSubmit() || (changeType === 'remove' && this.shouldNotAutoSubmitOnRemove())) {
+                return;
+            }
+
+            if (changeType === 'save' && 'terms' in data) {
+                // Replace old term data with the new one, as required by the backend
+                for (const termIndex of Object.keys(data['terms'])) {
+                    data['terms'][termIndex] = this.usedTerms[termIndex];
+                }
+            }
+
+            if (changeType === 'remove' && ! Object.keys(data['terms']).length) {
                 return;
             }
 
             this.dataInput.value = JSON.stringify({
                 type: changeType,
-                terms: changedTerms
+                ...data
             });
 
-            if (Object.keys(changedTerms).length) {
-                $(this.input.form).trigger('submit', { submittedBy: input });
+            let eventData = { submittedBy: input };
+            if (changeType === 'paste') {
+                // Ensure that what's pasted is also transmitted as value
+                eventData['terms'] = this.termsToQueryString(data['terms']) + this.separator + data['input'];
             }
+
+            $(this.input.form).trigger('submit', eventData);
         }
 
         submitTerms(terms) {
@@ -522,6 +640,10 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             );
         }
 
+        isTermDirectionVertical() {
+            return this.input.dataset.termDirection === 'vertical';
+        }
+
         moveFocusForward(from = null) {
             let toFocus;
 
@@ -533,6 +655,9 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
 
             if (from === -1) {
                 toFocus = inputs.shift();
+                if (typeof toFocus === 'undefined') {
+                    toFocus = this.input;
+                }
             } else if (from + 1 < inputs.length) {
                 toFocus = inputs[from + 1];
             } else {
@@ -613,12 +738,16 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             let termIndex = Number(input.parentNode.dataset.index);
 
             this.lastCompletedTerm = termData;
-            this.writePartialTerm(termData.label, input);
+
+            if ('label' in termData) {
+                this.writePartialTerm(termData.label, input);
+                this.checkValidity(input);
+            }
 
             if (termIndex >= 0) {
-                this.autoSubmit(input, 'save', { [termIndex]: this.saveTerm(input) });
+                this.autoSubmit(input, 'save', { terms: { [termIndex]: this.saveTerm(input, false, true) } });
             } else {
-                this.autoSubmit(input, 'exchange', this.exchangeTerm());
+                this.autoSubmit(input, 'exchange', { terms: this.exchangeTerm() });
                 this.togglePlaceholder();
             }
         }
@@ -635,11 +764,15 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             }
 
             if (! this.hasSyntaxError(input)) {
+                if (isTerm && ! this.validate(input)) {
+                    return;
+                }
+
                 this.complete(input, { term: termData });
             }
 
             if (! isTerm) {
-                this.autoSubmit(this.input, 'remove', this.clearSelectedTerms());
+                this.autoSubmit(this.input, 'remove', { terms: this.clearSelectedTerms() });
                 this.togglePlaceholder();
             }
         }
@@ -668,7 +801,9 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                 case 'Backspace':
                     removedTerms = this.clearSelectedTerms();
 
-                    if (termIndex >= 0 && ! input.value) {
+                    if (this.isTermDirectionVertical()) {
+                        // pass
+                    } else if (termIndex >= 0 && ! input.value) {
                         let removedTerm = this.removeTerm(input.parentNode);
                         if (removedTerm !== false) {
                             input = this.moveFocusBackward(termIndex);
@@ -695,12 +830,12 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                     }
 
                     this.togglePlaceholder();
-                    this.autoSubmit(input, 'remove', removedTerms);
+                    this.autoSubmit(input, 'remove', { terms: removedTerms });
                     break;
                 case 'Delete':
                     removedTerms = this.clearSelectedTerms();
 
-                    if (termIndex >= 0 && ! input.value) {
+                    if (! this.isTermDirectionVertical() && termIndex >= 0 && ! input.value) {
                         let removedTerm = this.removeTerm(input.parentNode);
                         if (removedTerm !== false) {
                             input = this.moveFocusForward(termIndex - 1);
@@ -716,11 +851,15 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                     }
 
                     this.togglePlaceholder();
-                    this.autoSubmit(input, 'remove', removedTerms);
+                    this.autoSubmit(input, 'remove', { terms: removedTerms });
                     break;
                 case 'Enter':
                     if (termIndex >= 0) {
-                        this.saveTerm(input, false);
+                        if (this.readPartialTerm(input)) {
+                            this.saveTerm(input, false);
+                        } else {
+                            this.removeTerm(input.parentNode, false);
+                        }
                     }
                     break;
                 case 'ArrowLeft':
@@ -735,6 +874,26 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
                         this.moveFocusForward();
                     }
                     break;
+                case 'ArrowUp':
+                    if (this.isTermDirectionVertical()
+                        && input.selectionStart === 0
+                        && this.hasTerms()
+                        && (this.completer === null || ! this.completer.isBeingCompleted(input))
+                    ) {
+                        event.preventDefault();
+                        this.moveFocusBackward();
+                    }
+                    break;
+                case 'ArrowDown':
+                    if (this.isTermDirectionVertical()
+                        && input.selectionStart === input.value.length
+                        && this.hasTerms()
+                        && (this.completer === null || ! this.completer.isBeingCompleted(input))
+                    ) {
+                        event.preventDefault();
+                        this.moveFocusForward();
+                    }
+                    break;
                 case 'a':
                     if ((event.ctrlKey || event.metaKey) && ! this.readPartialTerm(input)) {
                         this.selectTerms();
@@ -764,7 +923,7 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
 
                     break;
                 case 'Delete':
-                    this.autoSubmit(event.target, 'remove', this.clearSelectedTerms());
+                    this.autoSubmit(event.target, 'remove', { terms: this.clearSelectedTerms() });
                     this.togglePlaceholder();
                     break;
             }
@@ -781,15 +940,20 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             }
 
             // skipSaveOnBlur is set if the input is about to be removed anyway.
-            // If saveTerm would remove the input as well, the other removal will fail
-            // without any chance to handle it. (Element.remove() blurs the input)
+            // If we remove the input as well, the other removal will fail without
+            // any chance to handle it. (Element.remove() blurs the input)
             if (typeof input.skipSaveOnBlur === 'undefined' || ! input.skipSaveOnBlur) {
                 setTimeout(() => {
                     if (this.completer === null || ! this.completer.isBeingCompleted(input)) {
-                        let savedTerm = this.saveTerm(input);
-                        if (savedTerm !== false) {
-                            let termIndex = Number(input.parentNode.dataset.index);
-                            this.autoSubmit(input, 'save', { [termIndex]: savedTerm });
+                        let termIndex = Number(input.parentNode.dataset.index);
+                        if (this.readPartialTerm(input)) {
+                            let previousTerm = this.saveTerm(input);
+                            if (previousTerm !== false) {
+                                this.autoSubmit(input, 'save', { terms: { [termIndex]: previousTerm } });
+                            }
+                        } else {
+                            this.autoSubmit(
+                                input, 'remove', { terms: { [termIndex]: this.removeTerm(input.parentNode) } });
                         }
                     }
                 }, 0);
@@ -797,6 +961,12 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
         }
 
         onTermFocus(event) {
+            let input = event.target;
+
+            if (input.parentNode.dataset.index >= 0) {
+                this.validate(input);
+            }
+
             if (event.detail.scripted) {
                 // Only request suggestions if the user manually focuses the term
                 return;
@@ -804,8 +974,9 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
 
             this.deselectTerms();
 
-            let input = event.target;
-            if (! this.hasSyntaxError(input) && ! this.completer.isBeingCompleted(input, false)) {
+            if (! this.hasSyntaxError(input) && (
+                this.completer === null || ! this.completer.isBeingCompleted(input, false)
+            )) {
                 // Only request suggestions if the input is valid and not already being completed
                 let value = this.readPartialTerm(input);
                 this.complete(input, { trigger: 'script', term: { label: value } });
@@ -819,21 +990,30 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
             }
 
             if (this.hasTerms()) {
-                this.input.required = false;
+                if (this.manageRequired) {
+                    this.input.required = false;
+                }
 
                 // This is not part of `onSubmit()` because otherwise it would override what `autoSubmit()` does
                 this.dataInput.value = JSON.stringify({ type: 'submit', terms: this.usedTerms });
-            } else if (typeof this.input.dataset.manageRequired !== 'undefined') {
+
+                return;
+            } else if (this.manageRequired && ! this.hasTerms()) {
                 this.input.required = true;
             }
+
+            this.dataInput.value = '';
         }
 
         onPaste(event) {
-            if (this.hasTerms() || this.input.value) {
+            if (this.shouldNotAutoSubmit() || this.input.value) {
                 return;
             }
 
-            this.submitTerms(event.clipboardData.getData('text/plain'));
+            this.autoSubmit(this.input, 'paste', {
+                input: event.clipboardData.getData('text/plain'),
+                terms: this.usedTerms
+            });
 
             event.preventDefault();
         }
@@ -859,7 +1039,7 @@ define(["../notjQuery", "Completer"], function ($, Completer) {
 
             if (event.type === 'cut') {
                 this.clearPartialTerm(this.input);
-                this.autoSubmit(this.input, 'remove', this.clearSelectedTerms());
+                this.autoSubmit(this.input, 'remove', { terms: this.clearSelectedTerms() });
                 this.togglePlaceholder();
             }
         }
